/*
 * realmp.h
 *
 * Copyright (c) 2024, Jordi-Lluís Figueras, Alex Haro
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef __REALMP_H
#define __REALMP_H

#define REALMP_VERSION 1.0.0

#include<gmp.h>
#include<mpfr.h>
#include<iostream>

using namespace std;

class real
{
public:
  mpfr_t x;

  real();
  real(int a);
  real(unsigned int a);
  real(long a);
  real(unsigned long a);
  real(double a);
  real(const char *a);
  ~real();

  real& operator += (const real& y);

  real& operator = (const real& y);
  real& operator = (const double& y);
  real& operator = (const int& y);
};

inline real operator- (const real& P);

inline real operator+ (const real&, const real&);
inline real operator+ (const double&, const real&);
inline real operator+ (const real&, const double&);

inline real operator- (const real&, const real&);
inline real operator- (const double&, const real&);
inline real operator- (const real&, const double&);

inline real operator* (const real&, const real&);
inline real operator* (const double&, const real&);
inline real operator* (const real&, const double&);

inline real operator/ (const real&, const real&);
inline real operator/ (const double&, const real&);
inline real operator/ (const real&, const double&);

inline int operator> (const real& P, const real& Q);
inline int operator> (const real& P, double Q);
inline int operator> (double P, const real& Q);

inline int operator< (const real& P, const real& Q);
inline int operator< (const real& P, double Q);
inline int operator< (double P, const real& Q);

inline int operator>= (const real& P, const real& Q);
inline int operator>= (const real& P, double Q);
inline int operator>= (double P, const real& Q);

inline int operator<= (const real& P, const real& Q);
inline int operator<= (const real& P, double Q);
inline int operator<= (double P, const real& Q);

int prec_digits_to_bits(int digits);
void realmp_allocate_cache(int prec);
void realmp_free_cache(void);
inline real const_pi(void);
inline real sin(const real& P);
inline real cos(const real& P);
inline real acos(const real& P);
inline real tan(const real& P);
inline real atan(const real& P);
inline real atan2(const real& Q, const real& P);
inline real sinh(const real& P);
inline real cosh(const real& P);
inline real exp(const real& P);
inline real log(const real& P);
inline real log10(const real& P);
inline real sqr(const real& P);
inline real cubic(const real& P);
inline real sqrt(const real& P);
inline real abs(const real& P);
inline real pow(const real& P, int i);
inline real pow(const real& P, const real& x);
inline real max(const real& a, const real& b);
inline real max(const real& a, const real& b, const real& c);
inline real max(const real& a, const real& b, const real& c,
    const real& d);
inline real max(const real& a, const real& b, const real& c,
    const real& d, const real& e);
inline real ceil(const real& P);
inline real floor(const real& P);
real eval_pol(real* coef, const real& x, int N);
/*---------------------------------------------------------------------------*/
real::real () 
{
  mpfr_init(x);
  mpfr_set_d(x, (double)0., GMP_RNDN);
}

real::real (int a) 
{
  mpfr_init(x);
  // mpfr_set_d(x, (double)a, GMP_RNDN);
  mpfr_set_si(x, a, GMP_RNDN);
}

real::real (unsigned int a) 
{
  mpfr_init(x);
  // mpfr_set_d(x, (double)a, GMP_RNDN);
  mpfr_set_ui(x, a, GMP_RNDN);
}

real::real (long a) 
{
  mpfr_init(x);
  mpfr_set_si(x, a, GMP_RNDN);
}

real::real (unsigned long a) 
{
  mpfr_init(x);
  mpfr_set_ui(x, a, GMP_RNDN);
}


real::real (double a) 
{
  mpfr_init(x);
  mpfr_set_d(x, a, GMP_RNDN);
}

real::real (const char *a) 
{
  mpfr_init(x);
  mpfr_set_str(x, a, 10, GMP_RNDN);
}

real::~real()
{
  mpfr_clear(x);
}

real& real::operator += (const real& P)
{
  if(this != &P)
  {
      mpfr_add(x, x, P.x, GMP_RNDN);
  }
  return *this;
}


real& real::operator = (const real& P)
{
  if(this != &P)
  {
      mpfr_set(x, P.x, GMP_RNDN);
  }
  return *this;
}

real& real::operator = (const double& P)
{
  mpfr_set_d(x, P, GMP_RNDN);

  return *this;
}

real& real::operator = (const int& P)
{
  mpfr_set_si(x, P, GMP_RNDN);

  return *this;
}
 
// Global variables to be used everywhere
real val0, val1, val2, val3, val4, val5, val6, val7, val8, val9, val10;
real pi, pi2, pihalf;

   
#endif

