/*
 * realmpI.h
 *
 * Copyright (c) 2024, Jordi-Lluís Figueras, Alex Haro
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef __REALMPI_H
#define __REALMPI_H

#define REALMPI_VERSION 1.0

#include<mpfi.h>
#include<mpfi_io.h>
#include"realmp.h"

class realI
{
public:
	mpfi_t x;

  realI();
	realI(double a);
	realI(const real& a);
  realI(int a);
	realI(double a, double b);
	realI(const char *a);
	~realI();

	void setfull(double a, double b);
	void setfull(mpfr_t a, mpfr_t b);
	int position(double a);
	int position(mpfr_t a)const;
	int is_inside(double a);

	double left();
	double right();

	real left_real();
	real right_real();

	realI& operator = (const realI& y);
	realI& operator = (const double& y);
	realI& operator = (const mpfr_t& y);
	realI& operator = (const real& y);
  realI& operator += (const realI& P);
  realI& operator -= (const realI& P);
};

inline realI operator- (const realI& P);

inline realI operator+ (const realI&, const realI&);
inline realI operator+ (const double&, const realI&);
inline realI operator+ (const realI&, const double&);
inline realI operator+ (const mpfr_t&, const realI&);
inline realI operator+ (const realI&, const mpfr_t&);

inline realI operator- (const realI&, const realI&);
inline realI operator- (const double&, const realI&);
inline realI operator- (const realI&, const double&);
inline realI operator- (const mpfr_t&, const realI&);
inline realI operator- (const realI&, const mpfr_t&);
inline realI operator- (const real&, const realI&);
inline realI operator- (const realI&, const real&);

inline realI operator* (const realI&, const realI&);
inline realI operator* (const double&, const realI&);
inline realI operator* (const realI&, const double&);
inline realI operator* (const int&, const realI&);
inline realI operator* (const realI&, const int&);
inline realI operator* (const realI& P, const real& Q);
inline realI operator* (const real& Q, const realI& P);

inline realI operator/ (const realI&, const realI&);
inline realI operator/ (const double&, const realI&);
inline realI operator/ (const realI&, const double&);
inline realI operator/ (const realI&, const real&);

int is_inside(const realI& a, const realI& b);

void realmpI_allocate_cache(int prec);
inline realI midpointI(const realI& I);
inline realI const_piI(void);
inline realI abs(const realI& P);
inline realI sin(const realI& P);
inline realI sinh(const realI& P);
inline realI cos(const realI& P);
inline realI cosh(const realI& P);
inline realI exp(const realI& P);
inline realI log(const realI& P);
inline realI sqrt(const realI& P);
inline realI sqr(const realI& P);
inline realI atan2(const realI& P, const realI& Q);
inline realI pow(const realI& P, int N);

realI str_to_realI(char *s);

bool check_invertibility(realI **A, int dim);

realI eval_pol(realI* coef, const realI& x, int N);

/*-------------------------------------------------------------------------------------------------------------------------------------------*/

realI::realI () 
{
	mpfi_init_set_d(x, 0.);
}

realI::realI (double a) 
{
	mpfi_init_set_d(x, a);
}

realI::realI (const real& a) 
{
	mpfi_init_set_fr(x, a.x);
}

realI::realI (int a) 
{
	mpfi_init_set_si(x, a);
}

realI::realI (double a, double b) 
{
  mpfi_init(x);
	mpfi_interv_d(x, a, b);
}

realI::realI (const char *a) 
{
	mpfi_init_set_str(x, a, 10);
}

realI::~realI()
{
	mpfi_clear(x);
}

void realI::setfull(double a, double b)
{
	mpfi_interv_d(x, a, b);

	return ;
}

void realI::setfull(mpfr_t a, mpfr_t b)
{
	mpfi_interv_fr(x, a, b);

	return ;
}

int realI::position(double a)
{
	mpfr_t left, right;
	int flag = 0, retl, retr;

	mpfr_init(left);
	mpfr_init(right);

	mpfi_get_left(left, x);
	mpfi_get_right(right, x);

	retl = mpfr_cmp_d(left, a);
	retr = mpfr_cmp_d(right, a);
	
	if(retl > 0)
	{
		flag = -1;
	}
	if(retr < 0)
	{
		flag = 1;
	}

	mpfr_clear(left);
	mpfr_clear(right);

	return flag;
}

int realI::position(mpfr_t a)const
{
	mpfr_t left, right;
	int flag = 0, retl, retr;

	mpfr_init(left);
	mpfr_init(right);

	mpfi_get_left(left, x);
	mpfi_get_right(right, x);

	retl = mpfr_cmp(left, a);
	retr = mpfr_cmp(right, a);
	
	if(retl > 0)
	{
		flag = -1;
	}
	if(retr < 0)
	{
		flag = 1;
	}

	mpfr_clear(left);
	mpfr_clear(right);

	return flag;
}

//Returns 0 if a is not inside x, 1 otherwise
int realI::is_inside(double a)
{
	int flag;

	flag = mpfi_is_inside_d(a, x);

	return flag;
}

double realI::left(void )
{	
	mpfr_t left;
	double res;

	mpfr_init(left);

	mpfi_get_left(left, x);
	res = mpfr_get_d(left, GMP_RNDD);

	mpfr_clear(left);

	return res;
}

double realI::right(void )
{	
	mpfr_t right;
	double res;

	mpfr_init(right);

	mpfi_get_right(right, x);
	res = mpfr_get_d(right, GMP_RNDU);

	mpfr_clear(right);

	return res;
}

real realI::left_real(void )
{	
	real left;

	mpfi_get_left(left.x, x);

	return left;
}

real realI::right_real(void )
{	
	real right;

	mpfi_get_right(right.x, x);

	return right;
}

realI& realI::operator = (const realI& P)
{
	if(this != &P)
	{
		mpfi_set(x, P.x);
	}
	return *this;
}

realI& realI::operator = (const double& P)
{
	mpfi_set_d(x, P);
	return *this;
}

realI& realI::operator = (const mpfr_t& P)
{
	mpfi_set_fr(x, P);
	return *this;
}

realI& realI::operator = (const real& P)
{
	mpfi_set_fr(x, P.x);
	return *this;
}

realI& realI::operator += (const realI& P)
{
	mpfi_add(x, x, P.x);
	return *this;
}

realI& realI::operator -= (const realI& P)
{
	mpfi_sub(x, x, P.x);

	return *this;
}

#endif
